import logging
from pathlib import Path
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from ekorpkit.pipelines.pipe import apply_pipeline
from ekorpkit.io.file import load_dataframe
from hydra.utils import instantiate

log = logging.getLogger(__name__)


def build_corpus(**args):
    cfg = args.get(eKonf.Keys.CORPUS, {}).get("builtin", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


def build_t5(**args):
    cfg = args.get(eKonf.Keys.DATASET, {}).get("t5", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


def build_simple(**args):
    cfg = args.get(eKonf.Keys.DATASET, {}).get("simple", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


class DatasetBuilder:
    def __init__(self, **args) -> None:
        self.args = eKonf.to_dict(args)
        self.name = args.get("name", None)
        self.data_dir = args.get("data_dir", None)
        self.data_filetype = args.get("filetype", "parquet")
        self.column_info = self.args.get("column_info", None)
        self.verbose = self.args.get("verbose", False)
        self.autoload = self.args.get("autoload", False)

        self.io_args = args.get("io", None)
        self.fetch_dir = self.io_args.get("data_dir", None)
        self.fetch_sources = self.io_args.get("data_sources", None)
        if isinstance(self.fetch_sources, str):
            self.fetch_sources = [self.fetch_sources]
        if isinstance(self.fetch_sources, list):
            self.fetch_sources = {"train": self.fetch_sources}
        self.num_workers = self.io_args.get("num_workers", None)
        self.overwrite = self.io_args.get("overwrite", False)
        self.calculate_stats = self.io_args.get("calculate_stats", False)
        self.preprocess_text = self.io_args.get("preprocess_text", False)

        self.fetcher = self.io_args.get("fetcher", None)
        self.loader = self.io_args.get("loader", None)

        self.info_args = self.args.get("info", None)
        self.summary_info = None

        self.pipeline_args = self.args.get("pipeline", {})
        self.transform_pipeline = self.pipeline_args.get("_transform_", [])
        self.process_pipeline = self.pipeline_args.get("_preprocess_", [])
        if self.transform_pipeline is None:
            self.transform_pipeline = []
        if self.process_pipeline is None:
            self.process_pipeline = []

        if self.autoload:
            self.build()

    def build(self):
        if self.fetcher:
            if self.fetcher.get("_target_", None):
                eKonf.instantiate(self.fetcher)
            pipeline_args = self.fetcher.get("pipeline", None)
            if pipeline_args:
                eKonf.instantiate(pipeline_args)

        if self.info_args:
            self.summary_info = eKonf.instantiate(self.info_args)
        if self.summary_info:
            self.summary_info.load(self.args)

        for split_name, split_data_source in self.fetch_sources.items():
            if split_data_source is None:
                continue
            self._process_split(split_name)

        if self.summary_info:
            self.summary_info.save()

        log.info(
            f"\nCorpus [{self.name}] is built to [{self.data_dir}] from [{self.fetch_dir}]"
        )

    def _process_split(self, split_name):

        output_dir = Path(self.data_dir)
        if not output_dir.is_dir():
            output_dir.mkdir(exist_ok=True, parents=True)
        output_file = output_dir / f"{self.name}-{split_name}{self.data_filetype}"
        output_meta_file = (
            output_dir / f"meta-{self.name}-{split_name}{self.data_filetype}"
        )
        sample_file_prefix = f"{str(output_dir)}/sample-{self.name}-{split_name}"
        pipe = "save_metadata"
        if pipe in self.pipeline_args:
            if pipe not in self.transform_pipeline:
                self.transform_pipeline.append(pipe)
            self.pipeline_args[pipe]["filepath"] = str(output_meta_file)
            self.pipeline_args[pipe]["column_info"] = self.column_info
            self.pipeline_args[pipe]["split_name"] = split_name
        pipe = "save_samples"
        if pipe in self.pipeline_args:
            if pipe not in self.process_pipeline:
                self.process_pipeline.append(pipe)
            self.pipeline_args[pipe]["sample_file_prefix"] = sample_file_prefix
        pipe = "save_dataframe"
        if pipe in self.pipeline_args:
            if pipe not in self.process_pipeline:
                self.process_pipeline.append(pipe)
            self.pipeline_args[pipe]["filepath"] = str(output_file)
            columns = self.column_info.get("data")
            if columns:
                columns = list(columns.keys())
            self.pipeline_args[pipe]["columns"] = columns

        df = None
        if not output_file.exists() or self.overwrite:
            with elapsed_timer(format_time=True) as elapsed:
                df = instantiate(self.loader, split_name=split_name, _recursive_=False)
                log.info(f" >> elapsed time to load and parse data: {elapsed()}")

            if df is None:
                raise ValueError("dataframe is None")

            if self.verbose:
                print(df.head())
                print(df.shape)

            if self.transform_pipeline and len(self.transform_pipeline) > 0:
                log.info(
                    f"\nTransforming dataframe with pipeline: {self.transform_pipeline}"
                )
                df = apply_pipeline(df, self.transform_pipeline, self.pipeline_args)

            if self.summary_info and self.calculate_stats:
                stats = {
                    "name": split_name,
                    "dataset_name": self.name,
                    "data_file": output_file.name,
                }
                if output_meta_file.is_file():
                    stats["meta_file"] = output_meta_file.name
                self.summary_info.init_stats(df=df, split_name=split_name, stats=stats)

        else:
            log.info(f"{output_file} already exists")
            if self.calculate_stats or self.preprocess_text:
                df = load_dataframe(output_file, self.data_filetype)

        if df is None:
            log.warning("No datasets found")
            return None

        if self.process_pipeline and len(self.process_pipeline) > 0:
            log.info(f"\nProcessing dataframe with pipeline: {self.process_pipeline}")
            df = apply_pipeline(df, self.process_pipeline, self.pipeline_args)

        if self.calculate_stats and self.summary_info:
            self.summary_info.calculate_stats(df, split_name)



import logging
import os
import sklearn
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf
from ekorpkit.io.file import save_dataframe


log = logging.getLogger(__name__)


class SimpleTrainer:
    __metaclass__ = ABCMeta

    def __init__(self, **args):
        args = eKonf.to_dict(args)
        self.args = args
        self.name = args["name"]
        self.verbose = args.get("verbose", True)
        self._model_cfg = args["config"]
        self._model_eval = args.get("model", {}).get("eval")
        self._dataset = args.get(eKonf.Keys.DATASET, None)
        self._to_predict = args["to_predict"]
        self._to_train = args["to_train"]
        self._method_ = self.args.get("_method_")
        self._pred_output_dir = args["pred_output_dir"]
        self._pred_output_file = args["pred_output_file"]

        self._labels_key = self._to_train.get("labels")

        self._predicted_key = "predicted"
        self._model_outputs_key = "model_outputs"

        os.makedirs(args["output_dir"], exist_ok=True)
        os.makedirs(args["cache_dir"], exist_ok=True)
        os.makedirs(args["pred_output_dir"], exist_ok=True)
        os.makedirs(args["result_dir"], exist_ok=True)

        self.model = None
        self.dataset = None
        self.splits = {}
        self.train_data = None
        self.eval_data = None
        self.test_data = None

        eKonf.methods(self._method_, self)

    @abstractmethod
    def train(self):
        raise NotImplementedError("Must override train")

    @abstractmethod
    def _predict(self, to_predict: list):
        raise NotImplementedError("Must override predict")

    def load_datasets(self):
        if self._dataset is None:
            log.warning("No dataset config found")
            return
        self.dataset = eKonf.instantiate(self._dataset)
        self.splits = self.dataset.splits

        self.train_data = self.splits[self.dataset.SPLITS.TRAIN]
        if self.verbose:
            print("Train data:")
            print(self.train_data.info())
            print(self.train_data.tail())
        if "dev" in self.splits:
            self.eval_data = self.splits[self.dataset.SPLITS.DEV]
            self._model_cfg["evaluate_during_training"] = True
            if self.verbose:
                print("Eval data:")
                print(self.eval_data.info())
                print(self.eval_data.tail())
        else:
            self.eval_data = None
            self._model_cfg["evaluate_during_training"] = False
        self.test_data = self.splits[self.dataset.SPLITS.TEST]
        if self.verbose:
            print("Test data:")
            print(self.test_data.info())
            print(self.test_data.tail())

    def convert_to_predict(self, df):
        input_key = self._to_predict["input"]
        to_predict = df[input_key].tolist()
        if self.verbose:
            print(to_predict[:5])
        return to_predict

    def append_predictions(self, df, preds):
        predicted_column = self._to_predict[self._predicted_key]
        model_outputs_column = self._to_predict[self._model_outputs_key]
        df[predicted_column] = preds[self._predicted_key]
        df[model_outputs_column] = preds[self._model_outputs_key]
        return df

    def predict(self, df, _to_predict={}):
        if _to_predict:
            self._to_predict = _to_predict
        to_predict = self.convert_to_predict(df)
        preds = self._predict(to_predict)
        df = self.append_predictions(df, preds)
        return df

    def eval(self):
        if not self.splits:
            self.load_datasets()

        if self.test_data is None:
            log.warning("No test data found")
            return

        to_predict = self.convert_to_predict(self.test_data)
        preds = self._predict(to_predict)
        self.pred_data = self.append_predictions(self.test_data, preds)
        pred_filepath = os.path.join(self._pred_output_dir, self._pred_output_file)
        save_dataframe(self.pred_data, pred_filepath)
        if self.verbose:
            print(self.pred_data.head())
        if self._model_eval:
            eKonf.instantiate(self._model_eval, data=self.pred_data)


class SimpleNER(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.ner import NERModel

        args = self.args
        if args.labels is None:
            labels = list(self.train_data[self._labels_key].unique())

        # Create a NERModel
        model = NERModel(
            args.model_type,
            args.model_uri,
            labels=labels,
            cuda_device=args.cuda_device,
            args=eKonf.to_dict(args),
        )

        # Train the model
        model.train_model(self.train_data, eval_data=self.eval_data)

        # Evaluate the model
        result, model_outputs, predictions = model.eval_model(self.test_data)

        # Check predictions
        # print(predictions[:5])
        return result, model_outputs, predictions


class SimpleMultiLabel(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.classification import MultiLabelClassificationModel

        args = self.args
        # Create a Model
        model = MultiLabelClassificationModel(
            args.model_type,
            args.model_uri,
            num_labels=args.num_labels,
            args=eKonf.to_dict(args),
        )

        # Train the model
        model.train_model(self.train_data, eval_df=self.eval_data)

        # Evaluate the model
        result, model_outputs, predictions = model.eval_model(self.test_data)

        # Check predictions
        # print(predictions[:5])
        return result, model_outputs, predictions


class SimpleClassification(SimpleTrainer):
    def __init__(self, **args):
        super().__init__(**args)

    def train(self):
        from simpletransformers.classification import ClassificationModel

        args = self.args
        if not self.splits:
            self.load_datasets()

        self._model_cfg["labels_list"] = (
            self.train_data[self._labels_key].unique().tolist()
        )
        args["num_labels"] = len(self._model_cfg["labels_list"])

        # Create a NERModel
        model = ClassificationModel(
            args["model_type"],
            args["model_uri"],
            num_labels=args["num_labels"],
            cuda_device=args["cuda_device"],
            args=self._model_cfg,
        )

        # Train the model
        model.train_model(
            self.train_data, eval_df=self.eval_data, acc=sklearn.metrics.accuracy_score
        )

        # Evaluate the model
        result, model_outputs, wrong_predictions = model.eval_model(
            self.test_data, acc=sklearn.metrics.accuracy_score
        )
        if self.verbose:
            print(f"Evaluation result: {result}")
            print(f"Wrong predictions: {wrong_predictions[:5]}")
            # print(f"Model outputs: {model_outputs[:5]}")
            print(f"num_outputs: {len(model_outputs)}")
            print(f"num_wrong_predictions: {len(wrong_predictions)}")

    def load_model(self, model_dir=None):
        from simpletransformers.classification import ClassificationModel

        if model_dir is None:
            model_dir = self.args["best_model_dir"]

        self.model = ClassificationModel(
            self.args["model_type"], model_dir, args=self._model_cfg
        )
        log.info(f"Loaded model from {model_dir}")

    def _predict(self, to_predict: list):
        if self.model is None:
            self.load_model()

        predictions, raw_outputs = self.model.predict(to_predict)
        log.info(f"type of raw_outputs: {type(raw_outputs)}")
        raw_outputs= [output.flatten().tolist() for output in raw_outputs]
        log.info(f"raw_output: {raw_outputs[0]}")
        return {
            self._predicted_key: predictions,
            self._model_outputs_key: raw_outputs,
        }



import os
import logging
import pandas as pd
from collections import namedtuple
from tqdm import tqdm
from .score import get_process_memory, prune_vocab, NEG_INF
from ekorpkit.io.file import save_dataframe, load_dataframe
from ekorpkit import eKonf

# from inspect import getfullargspec as getargspec


log = logging.getLogger(__name__)


class Ngrams:
    def __init__(
        self,
        **args,
    ):
        args = eKonf.to_config(args)

        self.name = args.name
        self._data = args.data
        self.tokenize = args.tokenize

        self._ngram = args.ngram
        self._candidates = args.candidates
        self._surface = args.surface
        self._postag = args.postag
        self.score_function = eKonf.partial(args.score_function)

        assert type(self._ngram.max_n) == int

        if self._ngram.max_n <= 0:
            self._ngram.max_n = 4
        if self._ngram.max_window < self._ngram.max_n:
            self._ngram.max_window = self._ngram.max_n
        if not self._ngram.max_skip:
            self._ngram.max_skip = self._ngram.max_window - self._ngram.max_n

        if self._candidates.min_count <= 0:
            self._candidates.min_count = 10
        if self.tokenize is None:
            self.tokenize = lambda x: x.split()
        if self._candidates.max_candidates <= 0:
            self._candidates.max_candidates = 40000000
        self.progress_per = args.progress_per
        self.verbose = args.verbose

        self.output_dir = args.output_dir
        self.output_file = os.path.join(self.output_dir, args.output_file)
        os.makedirs(self.output_dir, exist_ok=True)

        self.sentences = []
        self.ngrams = {}
        self.candidates = {}
        self.total_words = 0

    def load_candidates(self):
        """Load a previously saved model"""
        if os.path.exists(self.output_file):
            df = load_dataframe(self.output_file, verbose=self.verbose)
            Ngram = namedtuple("ngram", df.columns)
            _cands = df.to_dict(orient="records")
            self.candidates = {
                tuple(cand["words"].split(self._ngram.delimiter)): Ngram(*cand)
                for cand in _cands
            }
            log.info(f"loaded {len(self.candidates)} candidates")
        else:
            log.info("no candidates to load")

    def save_candidates(self):
        """Save the candidates to a file"""
        if len(self.candidates) > 0:
            df = pd.DataFrame(self.candidates.values())
            save_dataframe(df, self.output_file, verbose=self.verbose)
        else:
            log.info("no candidates to save")

    def __str__(self):
        return (
            "%s<%i candidates, min_ngram_count=%s, threshold=%s, max_candidates=%s>"
            % (
                self.__class__.__name__,
                len(self.candidates),
                self._candidates.min_count,
                self._candidates.threshold,
                self._candidates.max_candidates,
            )
        )

    def _load_data(self):
        """Load data"""
        if self._data is None:
            log.warning("No data config found")
            return
        data = eKonf.instantiate(self._data)
        docs = data.data[data.COLUMN.TEXT]
        self.sentences = []
        for doc in docs:
            self.sentences.extend(doc.split("\n"))

    def train(self):
        """Train the model"""
        if not self.sentences:
            self._load_data()
        self._learn_ngrams()
        if self.score_function is not None and callable(self.score_function):
            self.candidates = self.score_function(self.ngrams)

    def _learn_ngrams(self):

        self.ngrams = {}
        sentence_no, total_words = -1, 0
        for sentence_no, sentence in tqdm(enumerate(self.sentences)):
            words = self.tokenize(sentence)
            total_words += len(words)

            ngrams = [
                tuple(words[pos : pos + n])
                for n in range(1, self._ngram.max_n + 1)
                for pos in range(0, len(words) - n + 1)
            ]
            ngrams = _prepare_ngram_tuples(
                words, max_n=self._ngram.max_n, max_window=self._ngram.max_window
            )
            for ngram in ngrams:
                self.ngrams[ngram] = self.ngrams.get(ngram, 0) + 1

            if (
                (sentence_no > 0)
                and (self._candidates.num_sents_per_pruning > 0)
                and (sentence_no % self._candidates.num_sents_per_pruning == 0)
            ):
                prune_vocab(self.ngrams, self._candidates.prune_min_ngram_count)
            if len(self.ngrams) > self._candidates.max_candidates:
                prune_vocab(self.ngrams, self._candidates.prune_min_ngram_count)
                self._candidates.prune_min_ngram_count += 1

            # if self.verbose and sentence_no % self.progress_per == 0:
            #     args = (
            #         len(self.ngrams),
            #         sentence_no + 1,
            #         len(self.sentences),
            #         get_process_memory(),
            #     )
            #     log.info(
            #         "learning ngrams ... # candidates= %d, (%d in %d) memory= %.3f Gb"
            #         % args
            #     )
        if self.verbose:
            print("learning ngrams was done. memory= %.3f Gb" % get_process_memory())

        self.ngrams = {
            ngram: freq
            for ngram, freq in self.ngrams.items()
            if freq >= self._candidates.min_count
        }
        self.total_words = total_words

    def find_ngrams(self, sentences):
        result = {}
        for sentence in sentences:
            for phrase, score in self.analyze_sentence(sentence):
                if score is not None:
                    result[phrase] = score
        return result

    def __getitem__(self, sentence):
        return [token for token, _ in self.analyze_sentence(sentence)]

    def concat_ngrams(self, sentence):
        return " ".join([token for token, _ in self.analyze_sentence(sentence)])

    def analyze_sentence(self, sentence, ngram_delim=None, strip_pos=None):
        """Analyze a sentence, concatenating any detected ngrams into a single token.

        Parameters
        ----------
        sentence : iterable of str
            Token sequence representing the sentence to be analyzed.

        Yields
        ------
            Iterate through the input sentence tokens and yield 2-tuples of:
            - ``(concatenated_ngram_tokens, score)`` for token sequences that form a phrase.
            - ``(word, None)`` if the token is not a part of a phrase.

        """

        start_token, in_between = None, []
        for word in sentence:
            if word not in self.connector_words:
                # The current word is a normal token, not a connector word, which means it's a potential
                # beginning (or end) of a phrase.
                if start_token:
                    # We're inside a potential phrase, of which this word is the end.
                    phrase, score = self.score_candidate(start_token, word, in_between)
                    if score is not None:
                        # Phrase detected!
                        yield phrase, score
                        start_token, in_between = None, []
                    else:
                        # Not a phrase after all. Dissolve the candidate's constituent tokens as individual words.
                        yield start_token, None
                        for w in in_between:
                            yield w, None
                        start_token, in_between = (
                            word,
                            [],
                        )  # new potential phrase starts here
                else:
                    # Not inside a phrase yet; start a new phrase candidate here.
                    start_token, in_between = word, []
            else:  # We're a connector word.
                if start_token:
                    # We're inside a potential phrase: add the connector word and keep growing the phrase.
                    in_between.append(word)
                else:
                    # Not inside a phrase: emit the connector word and move on.
                    yield word, None
        # Emit any non-phrase tokens at the end.
        if start_token:
            yield start_token, None
            for w in in_between:
                yield w, None

    def score_candidate(self, word_a, word_b, in_between):
        # Micro optimization: check for quick early-out conditions, before the actual scoring.
        word_a_cnt = self.candidates.get(word_a, 0)
        if word_a_cnt <= 0:
            return None, None

        word_b_cnt = self.candidates.get(word_b, 0)
        if word_b_cnt <= 0:
            return None, None

        phrase = self._ngram.delimiter.join([word_a] + in_between + [word_b])
        # XXX: Why do we care about *all* phrase tokens? Why not just score the start+end bigram?
        phrase_cnt = self.candidates.get(phrase, 0)
        if phrase_cnt <= 0:
            return None, None

        score = self.scoring(
            worda_count=word_a_cnt,
            wordb_count=word_b_cnt,
            bigram_count=phrase_cnt,
            len_vocab=len(self.candidates),
            min_count=self._candidates.min_count,
            corpus_word_count=self.corpus_word_count,
        )
        if score <= self._candidates.threshold:
            return None, None

        return phrase, score

    def export_ngrams(self, threshold=None):
        """Extract all found ngrams.
        Returns
        ------
        dict(str, float)
            Mapping between phrases and their scores.
        """
        result, source_vocab = {}, self.candidates
        for ngram in source_vocab:
            if len(ngram) < 2:
                continue  # no phrases here
            ngram, score = self.score_candidate(
                ngram[0], ngram[-1], ngram[1:-1]
            )
            if score is not None:
                result[ngram] = score
        return result


def _is_ordered_subset(subset, superset, distance_tolerance=None):
    if len(subset) > len(superset):
        return False
    if distance_tolerance and len(superset) - len(subset) > distance_tolerance:
        return False
    if len(subset) == len(superset):
        return subset == superset
    for sub_item in subset:
        if not superset:
            return False
        matched = -1
        for i, super_item in enumerate(superset):
            if sub_item == super_item:
                matched = i
                break
        if matched == -1:
            return False
        superset = superset[matched + 1 :]
    return True


def _prepare_ngram_tuples(words, max_n=5, max_window=None, max_skip=None):
    num_words = len(words)
    ngrams = [(word,) for word in words]
    ngrams += [
        tuple([words[i] for i in ngram_index_set])
        for ngram_index_set in _get_ngram_indices(
            num_words, max_n, max_window, max_skip
        )
    ]

    return ngrams


def _get_ngram_indices(num_words, max_n, max_window=None, max_skip=None):
    from itertools import combinations

    if max_window is None:
        max_window = max_n
    if max_skip is None:
        max_skip = 0
    if max_skip > max_n:
        max_skip = max_n

    word_positions = list(range(num_words))
    indices = set()
    for window in range(2, min(max_window, num_words) + 1):
        for starting in range(num_words - window + 1):
            position_set = word_positions[starting : starting + window]
            for n in range(max(2, window - max_skip), min(window, max_n) + 1):
                indices.update(set(combinations(position_set, n)))

    return sorted(indices)


def _get_surface(
    ngram,
    ngram_delim=";",
    surface_delim="",
    postag_delim="/",
    strip_pos=True,
    postag_length=None,
):
    tokens = ngram.split(ngram_delim)
    surfaces = [
        _get_word_surface(token, postag_delim, strip_pos, postag_length)
        for token in tokens
    ]
    return surface_delim.join(surfaces)


def _get_word_surface(token, postag_delim="/", strip_pos=True, postag_length=None):
    if not isinstance(postag_delim, str):
        return token
    token_pos = token.split(postag_delim)
    if strip_pos:
        return token_pos[0]
    return (
        token
        if len(token_pos) == 1
        else token_pos[0]
        + postag_delim
        + (token_pos[1][:postag_length] if postag_length else token_pos[1])
    )



import codecs
import logging
from abc import ABCMeta
from ekorpkit.io.load.list import load_wordlist
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Tokenizer:
    __metaclass__ = ABCMeta

    def __init__(
        self,
        normalize=None,
        tokenize={},
        tokenize_article={},
        extract={},
        verbose=False,
        **kwargs,
    ):

        self.verbose = verbose
        if tokenize is None:
            tokenize = {}
        if tokenize_article is None:
            tokenize_article = {}
        if extract is None:
            extract = {}

        self._normalize = normalize
        if eKonf.is_instantiatable(self._normalize):
            if self.verbose:
                log.info(f"instantiating {self._normalize['_target_']}...")
            self._normalize = eKonf.instantiate(self._normalize)

        self._lowercase = tokenize.get("lowercase", False)
        self._flatten = tokenize.get("flatten", True)
        self._concat_surface_and_pos = tokenize.get("concat_surface_and_pos", True)
        self._include_whitespace_token = tokenize.get("include_whitespace_token", True)
        self._tokenize_each_word = tokenize.get("tokenize_each_word", False)
        self._wordpieces_prefix = tokenize.get("wordpieces_prefix", "##")
        self._punct_postags = tokenize.get("punct_postags", None)
        if self._punct_postags is None:
            self._punct_postags = ["SF", "SP", "SSO", "SSC", "SY"]

        self._sentence_separator = tokenize_article.get("sentence_separator", None)
        if self._sentence_separator is None:
            self._sentence_separator = "\n"
        self._sentence_separator = codecs.decode(
            self._sentence_separator, "unicode_escape"
        )

        self._postags = extract.get("postags", None)
        if self._postags is None:
            self._postags = []
        self._noun_postags = extract.get("noun_postags", None)
        if self._noun_postags is None:
            self._noun_postags = ["NNG", "NNP", "XSN", "SL", "XR", "NNB", "NR"]
        self._stop_postags = extract.get("stop_postags", None)
        if self._stop_postags is None:
            self._stop_postags = ["SP"]
        self._no_space_for_non_nouns = extract.get("no_space_for_non_nouns", False)

        stopwords = kwargs.get("stopwords")
        self._stopwords = stopwords
        if eKonf.is_instantiatable(self._stopwords):
            if self.verbose:
                log.info(f"instantiating {self._stopwords['_target_']}...")
            self._stopwords = eKonf.instantiate(self._stopwords)

        self._return_as_list = kwargs.get("return_as_list", False)
        if self.verbose:
            log.info(f"{self.__class__.__name__} initialized with:")
            log.info(f"\treturn_as_list: {self._return_as_list}")

    def __call__(self, text):
        """Calling a tokenizer instance like a function just calls the tokenize method."""
        return self.tokenize(text)

    def parse(self, text):
        return text.split()

    def _to_token(self, term_pos):
        if isinstance(term_pos, tuple):
            if self._concat_surface_and_pos:
                return _tuple_to_token(term_pos)
            return term_pos[0]
        return term_pos

    def tokenize_article(self, article, return_as_list=None):
        if article is None:
            return None
        if return_as_list is None:
            return_as_list = self._return_as_list

        tokenized_article = []
        for sent in article.split(self._sentence_separator):
            sent = sent.strip()
            tokens = self.tokenize(sent, return_as_list=return_as_list)
            tokenized_article.append(tokens)
        return (
            tokenized_article
            if return_as_list
            else self._sentence_separator.join(tokenized_article)
        )

    def tokenize(self, text, return_as_list="list"):
        if isinstance(text, list):
            return text
        text = str(text)
        if self._lowercase:
            text = text.lower()
        if self._normalize and callable(self._normalize):
            text = self._normalize(text)
        if len(text) > 0:
            if self._tokenize_each_word:
                term_pos = []
                for word in text.split():
                    term_pos += self._tokenize_word(word)
            else:
                text = " ".join(text.split())
                term_pos = [self._to_token(token) for token in self.parse(text)]
        else:
            term_pos = []
        return term_pos if return_as_list else " ".join(term_pos)

    def pos(self, text):
        return self.tokenize(text)

    def _tokenize_word(self, word):
        tokens = self.parse(word)
        if len(tokens) > 1 and isinstance(tokens[0], tuple):
            term_pos = []
            for i, (term, pos) in enumerate(tokens):
                term = self._to_token((term, pos))
                if (
                    i > 0
                    and pos not in self._punct_postags
                    and tokens[i - 1][1] not in self._punct_postags
                ):
                    term = f"{self._wordpieces_prefix}{term}"
                term_pos.append(term)
            return term_pos
        return tokens

    def extract(self, text, nouns_only=False, return_as_list="list"):
        if nouns_only:
            tokens = _extract(
                text,
                postags=self._noun_postags,
                noun_postags=self._noun_postags,
                stopwords=self._stopwords,
            )
        else:
            tokens = _extract(
                text,
                postags=self._postags,
                stop_postags=self._stop_postags,
                no_space_for_non_nouns=self._no_space_for_non_nouns,
                stopwords=self._stopwords,
            )
        return tokens if return_as_list == "list" else " ".join(tokens)

    def extract_article(self, article, nouns_only=False):
        if article is None:
            return None

        tokens_article = []
        for sent in article.split(self._sentence_separator):
            sent = sent.strip()
            tokens = self.extract(
                sent, nouns_only=nouns_only, return_as_list=self._return_as_list
            )
            tokens_article.append(tokens)
        return (
            tokens_article
            if self._return_as_list
            else self._sentence_separator.join(tokens_article)
        )

    def extract_tokens(self, article):
        return self.extract_article(article, nouns_only=False)

    def extract_nouns(self, article):
        return self.extract_article(article, nouns_only=True)

    def filter_stopwords(self, text_or_tokens):
        if text_or_tokens is None:
            return None

        if isinstance(text_or_tokens, list):
            tokens = text_or_tokens
        else:
            tokens = self.tokenize(text_or_tokens)

        tokens = [token for token in tokens if not self._stopwords(token)]

        return tokens if self._return_as_list else " ".join(tokens)

    def filter_article_stopwords(self, article):
        if article is None:
            return None

        token_article = []
        for sent in article.split(self._sentence_separator):
            token_article.append(self.filter_stopwords(sent))
        return (
            token_article
            if self._return_as_list
            else self._sentence_separator.join(token_article)
        )

    def nouns(self, text):
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=True)

    def tokens(self, text):
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=False)

    def morphs(self, text):
        return self.tokens(text)


def _tuple_to_token(_tuple):
    if isinstance(_tuple, tuple):
        return "/".join(list(_tuple))
    return _tuple


def _token_to_tuple(_token):
    if isinstance(_token, str):
        return tuple(_token.split("/"))
    return _token


def _extract(
    tokenized_text,
    postags=[],
    noun_postags=["NNG", "NNP", "XSN", "SL", "XR", "NNB", "NR"],
    stop_postags=["SP"],
    no_space_for_non_nouns=False,
    stopwords=None,
    **kwargs,
):
    if isinstance(tokenized_text, str):
        tokens = tokenized_text.split()
    else:
        tokens = tokenized_text
    _tokens_pos = [
        _token_to_tuple(token) for token in tokens if len(_token_to_tuple(token)) == 2
    ]

    exist_sp_tag = False
    if no_space_for_non_nouns:
        for i, token in enumerate(_tokens_pos):
            if token[1] == "SP":
                exist_sp_tag = True
                break

    _tokens = []
    if exist_sp_tag and no_space_for_non_nouns:
        prev_nonnoun_check = False
        cont_morphs = []
        i = 0
        while i < len(_tokens_pos):
            token = _tokens_pos[i]
            if not prev_nonnoun_check and token[1] in noun_postags:
                _tokens.append(token[0])
            elif (
                not prev_nonnoun_check
                and token[1] not in noun_postags
                and token[1][0] != "S"
            ):
                prev_nonnoun_check = True
                cont_morphs.append(token[0])
            elif (
                prev_nonnoun_check
                and token[1] not in noun_postags
                and token[1][0] != "S"
            ):
                cont_morphs.append(token[0])
            else:
                if len(cont_morphs) > 0:
                    _tokens.append("".join(cont_morphs))
                    cont_morphs = []
                    prev_nonnoun_check = False
                if token[1] != "SP":
                    _tokens.append(token[0])
            i += 1
        if len(cont_morphs) > 0:
            _tokens.append("".join(cont_morphs))
    else:
        if len(postags) > 0:
            _tokens = [
                token[0].strip()
                for token in _tokens_pos
                if token[1] not in stop_postags and token[1] in postags
            ]
        else:
            _tokens = [
                token[0].strip()
                for token in _tokens_pos
                if token[1] not in stop_postags
            ]

    if stopwords is not None:
        _tokens = [token for token in _tokens if not stopwords(token)]
    return _tokens


class NLTKTokenizer(Tokenizer):
    def __init__(self, nltk={}, **kwargs):
        import nltk as NLTK

        NLTK.download("punkt", quiet=True)
        NLTK.download("averaged_perceptron_tagger", quiet=True)
        NLTK.download('wordnet', quiet=True)
        NLTK.download('omw-1.4', quiet=True)

        nltk = nltk or {}
        self.verbose = kwargs.get("verbose", False)

        self.lemmatizer = nltk.get("lemmatizer", None)
        if eKonf.is_instantiatable(self.lemmatizer):
            if self.verbose:
                log.info(f"instantiating {self.lemmatizer['_target_']}...")
            self.lemmatizer = eKonf.instantiate(self.lemmatizer)
        self.stemmer = nltk.get("stemmer", None)
        if eKonf.is_instantiatable(self.stemmer):
            if self.verbose:
                log.info(f"instantiating {self.stemmer['_target_']}...")
            self.stemmer = eKonf.instantiate(self.stemmer)
        do_lemmatize = nltk.get("lemmatize", False)
        do_stem = nltk.get("stem", False)
        self.do_lemmatize = do_lemmatize and self.lemmatizer is not None
        self.do_stem = do_stem and self.stemmer is not None and not self.do_lemmatize

        super().__init__(**kwargs)

    def parse(self, text):
        import nltk

        tokens = nltk.pos_tag(nltk.word_tokenize(text))
        if self.do_lemmatize:
            tokens = self.lemmatize(tokens)
        if self.do_stem:
            tokens = self.stem(tokens)

        return tokens

    def _lemmatize(self, token):
        if self.lemmatizer is None:
            return token
        token_pos = _token_to_tuple(token)
        if len(token_pos) == 2:
            return (
                self.lemmatizer.lemmatize(
                    token_pos[0], self._get_wordnet_pos(token_pos[1])
                ),
                token_pos[1],
            )
        return self.lemmatizer.lemmatize(token_pos[0])

    def _stem(self, token):
        if self.stemmer is None:
            return token
        token_pos = _token_to_tuple(token)
        if len(token_pos) == 2:
            return (self.stemmer.stem(token_pos[0]), token_pos[1])
        return self.stemmer.stem(token_pos[0])

    def lemmatize(self, tokens):
        tokens = [self._lemmatize(token) for token in tokens]
        return tokens

    def stem(self, tokens):
        tokens = [self._stem(token) for token in tokens]
        return tokens

    @staticmethod
    def _get_wordnet_pos(tag):
        from nltk.corpus import wordnet

        """Map POS tag to first character lemmatize() accepts"""
        tag = tag[0].upper()
        tag_dict = {
            "J": wordnet.ADJ,
            "N": wordnet.NOUN,
            "V": wordnet.VERB,
            "R": wordnet.ADV,
        }

        return tag_dict.get(tag, wordnet.NOUN)


class SimpleTokenizer(Tokenizer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse(self, text):
        return text.split()


class PynoriTokenizer(Tokenizer):
    def __init__(
        self,
        pynori={},
        **kwargs,
    ):
        log.info(f"Initializing Pynori with {pynori}...")

        super().__init__(**kwargs)
        if pynori is None:
            pynori = {}
        pynori["decompound_mode"] = "NONE" if self._flatten else "MIXED"
        pynori["infl_decompound_mode"] = "NONE" if self._flatten else "MIXED"
        pynori["discard_punctuation"] = False
        if pynori["path_userdict"] is None:
            pynori.pop("path_userdict")

        try:
            from pynori.korean_analyzer import KoreanAnalyzer

            if pynori is None:
                self._tokenizer = KoreanAnalyzer()
            else:
                self._tokenizer = KoreanAnalyzer(**pynori)
        except ImportError:
            raise ImportError(
                "\n"
                "You must install `pynori` if you want to use `pynori` backend.\n"
                "Please install using `pip install pynori`.\n"
            )

    def parse(self, text):

        _concat_surface_and_pos = (
            False if self._tokenize_each_word else self._concat_surface_and_pos
        )
        tokens = self._tokenizer.do_analysis(text)
        term_pos = [
            (term, pos)
            for term, pos in zip(tokens["termAtt"], tokens["posTagAtt"])
            if self._include_whitespace_token or pos != "SP"
        ]
        return term_pos


class MecabTokenizer(Tokenizer):
    def __init__(
        self,
        mecab={},
        **kwargs,
    ):

        log.info(f"Initializing mecab with {mecab}...")
        super().__init__(**kwargs)
        self.mecab = mecab
        try:
            from ..models.tokenizer.mecab import MeCab

            if self.mecab is None:
                _tokenizer = MeCab()
            else:
                _tokenizer = MeCab(**self.mecab)
        except ImportError:
            raise ImportError(
                "\n"
                "You must install `fugashi` and `mecab_ko_dic` if you want to use `mecab` backend.\n"
                "Please install using `pip install python-mecab-ko`.\n"
            )

    def parse(self, text):
        from ..models.tokenizer.mecab import MeCab

        if self.mecab is None:
            self._tokenizer = MeCab()
        else:
            self._tokenizer = MeCab(**self.mecab)

        return self._tokenizer.pos(
            text,
            concat_surface_and_pos=False,
            flatten=self._flatten,
            include_whitespace_token=self._include_whitespace_token,
        )


class BWPTokenizer(Tokenizer):
    def __init__(
        self,
        bwp={},
        **kwargs,
    ):
        log.info("Initializing BertWordPieceTokenizer...")
        super().__init__(**kwargs)
        try:
            from transformers import BertTokenizerFast

            if bwp is None or bwp.get("pretrained_model_name_or_path") is None:
                self._tokenizer = BertTokenizerFast.from_pretrained(
                    "entelecheia/ekonbert-base"
                )
            else:
                self._tokenizer = BertTokenizerFast.from_pretrained(**bwp)

        except ImportError:
            raise ImportError(
                "\n"
                "You must install `BertWordPieceTokenizer` if you want to use `bwp` backend.\n"
                "Please install using `pip install transformers`.\n"
            )

    def parse(self, text):
        return self._tokenizer.tokenize(text)



import shutil
from pathlib import Path
from ekorpkit import eKonf


def utils(**args):
    args = eKonf.to_config(args)
    subtask = args.subtask

    if subtask == "duplicate":
        duplicate_folder(args)


def duplicate_folder(args):
    dupe_factor = args.dupe_factor
    input_dir = Path(args.input_dir)
    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True)

    if dupe_factor > 1:
        if input_dir.is_dir():
            src = str(input_dir)
            print("Source folder: ", src)
            for i in range(dupe_factor):
                if i == 0:
                    continue
                # Destination path
                dest = output_dir / f"{input_dir.name}_{i}"
                if dest.is_dir():
                    print("Destination folder already exists, skipping.")
                else:
                    destination = shutil.copytree(src, dest)
                    print("Duplicated folder as the path:", destination)
        else:
            print("input directory does not exist")

    else:
        print("dupe_factor must be greater than 1")



import codecs
import logging
import pandas as pd
from ekorpkit import eKonf

log = logging.getLogger(__name__)


class ColumnInfo:
    def __init__(self, **args):
        self.args = eKonf.to_dict(args)

        self._merge_meta_on_key = "merge_meta_on"
        self._orginal_id_key = "_id"

        self.KEYs = self.KEYs
        self.COLUMNs = self.COLUMNs
        self.DATATYPEs = self.DATATYPEs
        self.METATYPEs = self.METATYPEs
        self.TIMESTAMP_PARM = self.TIMESTAMP_PARM
        self.DATETIME_PARM = self.DATETIME_PARM

        self.ID = self.ID
        self._ID = self._ID
        self.TEXT = self.TEXT
        self.TIMESTAMP = self.TIMESTAMP
        self.SPLIT = self.SPLIT
        self.CORPUS = self.CORPUS
        self.DATASET = self.DATASET

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname} : {self.INFO}"
        return s

    def to_timestamp(self, data, metadata=None):
        if self.TIMESTAMP_PARM is None:
            return data, metadata

        _key = self.TIMESTAMP_PARM.get("key")
        _format = self.TIMESTAMP_PARM.get("format")
        _parms_ = self.TIMESTAMP_PARM.get(eKonf.Keys.PARMS) or {}
        if _key is None:
            log.info("No timestamp key found")
            return data, metadata
        if isinstance(data, pd.DataFrame):
            if _key in data.columns:
                data[self.TIMESTAMP] = pd.to_datetime(
                    data[_key], format=_format, **_parms_
                )
                log.info(f"Loaded timestamp column {self.TIMESTAMP}")
            elif metadata is not None and _key in metadata.columns:
                metadata[self.TIMESTAMP] = pd.to_datetime(
                    metadata[_key], format=_format, **_parms_
                )
                df_dt = metadata[self.MERGE_META_ON + [self.TIMESTAMP]].copy()
                data = data.merge(df_dt, on=self.MERGE_META_ON, how="left")
                metadata.drop(self.TIMESTAMP, axis=1, inplace=True)
                log.info(f"Timestamp column {self.TIMESTAMP} added to data")
        return data, metadata

    def to_datetime(self, data):
        if self.DATETIME_PARM is None:
            return data

        _columns = eKonf.ensure_list(self.DATETIME_PARM.get("key"))
        _format = self.DATETIME_PARM.get("format", None)
        _parms_ = self.DATETIME_PARM.get(eKonf.Keys.PARMS) or {}
        if _columns is None:
            log.info("No datetime column found")
            return data
        if isinstance(data, pd.DataFrame):
            for _col in _columns:
                if _col in data.columns:
                    data[_col] = pd.to_datetime(data[_col], format=_format, **_parms_)
                    log.info(f"converted datetime column {_col}")
        return data

    def combine_texts(self, data):
        if self.TEXTs is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.TEXTs] = data[self.TEXTs].fillna("")
            if len(self.TEXTs) > 1:
                data[self.TEXT] = data[self.TEXTs].apply(
                    lambda row: self.SEGMENT_SEP.join(row.values.astype(str)),
                    axis=1,
                )
                self.DATATYPEs = {
                    k: v for k, v in self.DATATYPEs.items() if k not in self.TEXTs
                }
                self.DATATYPEs[self.TEXT] = "str"

        return data

    def combine_ids(self, data):
        if self.IDs is None:
            return data

        if isinstance(data, pd.DataFrame):
            if len(self.IDS) > 1:
                data[self.ID] = data[self.IDs].apply(
                    lambda row: self.ID_SEPARATOR.join(row.values.astype(str)),
                    axis=1,
                )

        return data

    def merge_metadata(self, data, metadata):
        if metadata is None:
            return data
        data = data.merge(
            metadata,
            on=self.MERGE_META_ON,
            how="left",
        )
        return data

    def append_split(self, data, _split):
        if _split is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.SPLIT] = _split
            if self.SPLIT not in self.IDs:
                self.IDs.append(self.SPLIT)
            if self.DATA and self.SPLIT not in self.DATA:
                self.DATATYPEs[self.SPLIT] = "str"
            if self.METADATA and self.SPLIT not in self.METADATA:
                self.METATYPEs[self.SPLIT] = "str"

            log.info(f"Added split column [{self.SPLIT}] with value [{_split}]")

        return data

    def append_corpus(self, data, _corpus):
        if _corpus is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.CORPUS] = _corpus
            if self.CORPUS not in self.IDs:
                self.IDs.append(self.CORPUS)
            if self.DATA and self.CORPUS not in self.DATA:
                self.DATATYPEs[self.CORPUS] = "str"
            if self.METADATA and self.CORPUS not in self.METADATA:
                self.METATYPEs[self.CORPUS] = "str"

            log.info(f"Added corpus column [{self.CORPUS}] with value [{_corpus}]")

        return data

    def common_columns(self, dataframes):
        """
        Find common columns between dataframes
        """
        if not isinstance(dataframes, list):
            dataframes = [dataframes]
        common_columns = list(set.intersection(*(set(df.columns) for df in dataframes)))
        df = dataframes[0][common_columns].copy()
        dtypes = df.dtypes.apply(lambda x: x.name).to_dict()
        self.DATATYPEs = dtypes
        return common_columns

    def append_dataset(self, data, _dataset):
        if _dataset is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.DATASET] = _dataset
            if self.DATASET not in self.IDs:
                self.IDs.append(self.DATASET)
            if self.DATA and self.DATASET not in self.DATA:
                self.DATATYPEs[self.DATASET] = "str"

            log.info(f"Added dataset column [{self.DATASET}] with value [{_dataset}]")

        return data

    def reset_id(self, df):
        if isinstance(df, pd.DataFrame):
            df.rename({self.ID: self._ID}, inplace=True)
            df.reset_index().rename({"index": self.ID}, inplace=True)
            dtypes = df.dtypes.apply(lambda x: x.name).to_dict()
            self.DATATYPEs = dtypes
        return df

    @property
    def INFO(self):
        return self.args

    @property
    def ID(self):
        return eKonf.Keys.ID

    @ID.setter
    def ID(self, value):
        self.KEYs[eKonf.Keys.ID] = value

    @property
    def _ID(self):
        return self.KEYs.get(self._orginal_id_key) or self._orginal_id_key

    @_ID.setter
    def _ID(self, value):
        self.KEYs[self._orginal_id_key] = value

    @property
    def ID_SEPARATOR(self):
        return "_"

    @property
    def MERGE_META_ON(self):
        return eKonf.ensure_list(self.COLUMNs[self._merge_meta_on_key]) or self.IDs

    @MERGE_META_ON.setter
    def MERGE_META_ON(self, value):
        self.COLUMNs[self._merge_meta_on_key] = value

    @property
    def IDs(self):
        return eKonf.ensure_list(self.COLUMNs[eKonf.Keys.ID])

    @IDs.setter
    def IDs(self, value):
        self.COLUMNs[eKonf.Keys.ID] = value

    @property
    def TEXT(self):
        return eKonf.Keys.TEXT

    @TEXT.setter
    def TEXT(self, value):
        self.KEYs[eKonf.Keys.TEXT] = value

    @property
    def TEXTs(self):
        return eKonf.ensure_list(self.COLUMNs[eKonf.Keys.TEXT])

    @TEXTs.setter
    def TEXTs(self, value):
        self.COLUMNs[eKonf.Keys.TEXT] = value

    @property
    def DATA(self):
        if self.DATATYPEs is None:
            return None
        return list(self.DATATYPEs.keys())

    @property
    def METADATA(self):
        if self.METATYPEs is None:
            return None
        return list(self.METATYPEs.keys())

    @property
    def TIMESTAMP(self):
        return eKonf.Keys.TIMESTAMP

    @TIMESTAMP.setter
    def TIMESTAMP(self, value):
        self.KEYs[eKonf.Keys.TIMESTAMP] = value

    @property
    def SPLIT(self):
        return self.KEYs.get(eKonf.Keys.SPLIT) or eKonf.Keys.SPLIT

    @SPLIT.setter
    def SPLIT(self, value):
        self.KEYs[eKonf.Keys.SPLIT] = value

    @property
    def CORPUS(self):
        return self.KEYs.get(eKonf.Keys.CORPUS) or eKonf.Keys.CORPUS

    @CORPUS.setter
    def CORPUS(self, value):
        self.KEYs[eKonf.Keys.CORPUS] = value

    @property
    def DATASET(self):
        return self.KEYs.get(eKonf.Keys.DATASET) or eKonf.Keys.DATASET

    @DATASET.setter
    def DATASET(self, value):
        self.KEYs[eKonf.Keys.DATASET] = value

    @property
    def KEYs(self):
        return self.INFO.get("keys") or {}

    @KEYs.setter
    def KEYs(self, value):
        self.INFO["keys"] = value

    @property
    def COLUMNs(self):
        return self.INFO.get("columns") or {}

    @COLUMNs.setter
    def COLUMNs(self, value):
        self.INFO["columns"] = value

    @property
    def DATATYPEs(self):
        return self.INFO.get("data")

    @DATATYPEs.setter
    def DATATYPEs(self, value):
        self.INFO["data"] = value

    @property
    def METATYPEs(self):
        return self.INFO.get("meta")

    @METATYPEs.setter
    def METATYPEs(self, value):
        self.INFO["meta"] = value

    @property
    def TIMESTAMP_PARM(self):
        return self.INFO.get(eKonf.Keys.TIMESTAMP)

    @TIMESTAMP_PARM.setter
    def TIMESTAMP_PARM(self, value):
        self.INFO[eKonf.Keys.TIMESTAMP] = value

    @property
    def DATETIME_PARM(self):
        return self.INFO.get(eKonf.Keys.DATETIME)

    @DATETIME_PARM.setter
    def DATETIME_PARM(self, value):
        self.INFO[eKonf.Keys.DATETIME] = value

    @property
    def SEGMENT_SEP(self):
        return codecs.decode(
            self.INFO.get("segment_separator", "\n\n"), "unicode_escape"
        )

    @property
    def SENTENCE_SEP(self):
        return codecs.decode(
            self.INFO.get("sentence_separator", "\n"), "unicode_escape"
        )
